@@ -1,5 +1,5 @@
 /****************************************************************************
- * Copyright (c) 1998-2000,2003 Free Software Foundation, Inc.              *
+ * Copyright (c) 1998-2003,2004 Free Software Foundation, Inc.              *
  *                                                                          *
  * Permission is hereby granted, free of charge, to any person obtaining a  *
  * copy of this software and associated documentation files (the            *
@@ -28,10 +28,9 @@
 
 /****************************************************************************
  *   Author:  Juergen Pfeifer, 1995,1997                                    *
- *   Contact: http://www.familiepfeifer.de/Contact.aspx?Lang=en             *
  ****************************************************************************/
 
-/* $Id: form.h,v 0.15 2003/05/03 21:59:32 tom Exp $ */
+/* $Id: form.h,v 0.19 2004/05/15 21:00:52 tom Exp $ */
 
 #ifndef FORM_H
 #define FORM_H
@@ -43,6 +42,14 @@
   extern "C" {
 #endif
 
+#ifndef FORM_PRIV_H
+typedef void *FIELD_CELL;
+#endif
+
+#ifndef NCURSES_FIELD_INTERNALS
+#define NCURSES_FIELD_INTERNALS /* nothing */
+#endif
+
 typedef int Form_Options;
 typedef int Field_Options;
 
@@ -51,10 +58,10 @@ typedef int Field_Options;
 	**********/
 
 typedef struct {
-  short	pmin;	  /* index of first field on page            */
-  short	pmax;	  /* index of last field on page             */
-  short	smin;	  /* index of top leftmost field on page     */
-  short	smax;	  /* index of bottom rightmost field on page */
+  short pmin;		/* index of first field on page			*/
+  short pmax;		/* index of last field on page			*/
+  short smin;		/* index of top leftmost field on page		*/
+  short smax;		/* index of bottom rightmost field on page	*/
 } _PAGE;
 
 	/**********
@@ -62,31 +69,40 @@ typedef struct {
 	**********/
 
 typedef struct fieldnode {
-  unsigned short       	status;	  /* flags		        */
-  short			rows;	  /* size in rows		*/
-  short			cols;	  /* size in cols		*/
-  short			frow;	  /* first row		        */
-  short			fcol;	  /* first col		        */
-  int                   drows;    /* dynamic rows               */
-  int                   dcols;    /* dynamic cols               */
-  int                   maxgrow;  /* maximum field growth       */
-  int			nrow;	  /* offscreen rows	        */
-  short			nbuf;	  /* additional buffers	        */
-  short			just;	  /* justification	        */
-  short			page;	  /* page on form		*/
-  short			index;	  /* into form -> field	        */
-  int			pad;	  /* pad character	        */
-  chtype		fore;	  /* foreground attribute	*/
-  chtype		back;	  /* background attribute	*/
-  Field_Options		opts;	  /* options		        */
-  struct fieldnode *	snext;	  /* sorted order pointer	*/
-  struct fieldnode *	sprev;	  /* sorted order pointer	*/
-  struct fieldnode *	link;	  /* linked field chain	        */
-  struct formnode *	form;	  /* containing form	        */
-  struct typenode *	type;	  /* field type		        */
-  void *		arg;	  /* argument for type	        */
-  char *		buf;	  /* field buffers	        */
-  void *		usrptr;	  /* user pointer		*/
+  unsigned short	status;		/* flags			*/
+  short			rows;		/* size in rows			*/
+  short			cols;		/* size in cols			*/
+  short			frow;		/* first row			*/
+  short			fcol;		/* first col			*/
+  int			drows;		/* dynamic rows			*/
+  int			dcols;		/* dynamic cols			*/
+  int			maxgrow;	/* maximum field growth		*/
+  int			nrow;		/* off-screen rows		*/
+  short			nbuf;		/* additional buffers		*/
+  short			just;		/* justification		*/
+  short			page;		/* page on form			*/
+  short			index;		/* into form -> field		*/
+  int			pad;		/* pad character		*/
+  chtype		fore;		/* foreground attribute		*/
+  chtype		back;		/* background attribute		*/
+  Field_Options		opts;		/* options			*/
+  struct fieldnode *	snext;		/* sorted order pointer		*/
+  struct fieldnode *	sprev;		/* sorted order pointer		*/
+  struct fieldnode *	link;		/* linked field chain		*/
+  struct formnode *	form;		/* containing form		*/
+  struct typenode *	type;		/* field type			*/
+  void *		arg;		/* argument for type		*/
+  FIELD_CELL *		buf;		/* field buffers		*/
+  void *		usrptr;		/* user pointer			*/
+  /*
+   * The wide-character configuration requires extra information.  Because
+   * there are existing applications that manipulate the members of FIELD
+   * directly, we cannot make the struct opaque.  Offsets of members up to
+   * this point are the same in the narrow- and wide-character configuration.
+   * But note that the type of buf depends on the configuration, and is made
+   * opaque for that reason.
+   */
+  NCURSES_FIELD_INTERNALS
 } FIELD;
 
 	/**************
@@ -94,20 +110,20 @@ typedef struct fieldnode {
 	**************/
 
 typedef struct typenode {
-  unsigned short	status;	               /* flags		       */
-  long			ref;	               /* reference count      */
-  struct typenode *	left;	               /* ptr to operand for | */
-  struct typenode *	right;	               /* ptr to operand for | */
+  unsigned short	status;			/* flags		*/
+  long			ref;			/* reference count	*/
+  struct typenode *	left;			/* ptr to operand for | */
+  struct typenode *	right;			/* ptr to operand for | */
 
-  void* (*makearg)(va_list *);                 /* make fieldtype arg   */
-  void* (*copyarg)(const void *);              /* copy fieldtype arg   */
-  void  (*freearg)(void *);                    /* free fieldtype arg   */
+  void* (*makearg)(va_list *);			/* make fieldtype arg	*/
+  void* (*copyarg)(const void *);		/* copy fieldtype arg	*/
+  void	(*freearg)(void *);			/* free fieldtype arg	*/
 
-  bool	(*fcheck)(FIELD *,const void *);       /* field validation     */
-  bool	(*ccheck)(int,const void *);           /* character validation */
+  bool	(*fcheck)(FIELD *,const void *);	/* field validation	*/
+  bool	(*ccheck)(int,const void *);		/* character validation */
 
-  bool	(*next)(FIELD *,const void *);         /* enumerate next value */
-  bool	(*prev)(FIELD *,const void *);         /* enumerate prev value */
+  bool	(*next)(FIELD *,const void *);		/* enumerate next value */
+  bool	(*prev)(FIELD *,const void *);		/* enumerate prev value */
 
 } FIELDTYPE;
 
@@ -116,29 +132,29 @@ typedef struct typenode {
 	*********/
 
 typedef struct formnode {
-  unsigned short	status;	  /* flags		        */
-  short			rows;	  /* size in rows		*/
-  short			cols;	  /* size in cols		*/
-  int			currow;	  /* current row in field window*/
-  int			curcol;	  /* current col in field window*/
-  int			toprow;	  /* in scrollable field window	*/
-  int                   begincol; /* in horiz. scrollable field */
-  short			maxfield; /* number of fields	        */
-  short			maxpage;  /* number of pages	        */
-  short			curpage;  /* index into page	        */
-  Form_Options		opts;	  /* options		        */
-  WINDOW *		win;	  /* window		        */
-  WINDOW *		sub;	  /* subwindow		        */
-  WINDOW *		w;	  /* window for current field	*/
-  FIELD **		field;	  /* field [maxfield]	        */
-  FIELD *		current;  /* current field	        */
-  _PAGE *		page;	  /* page [maxpage]	        */
-  void *		usrptr;	  /* user pointer		*/
+  unsigned short	status;		/* flags			*/
+  short			rows;		/* size in rows			*/
+  short			cols;		/* size in cols			*/
+  int			currow;		/* current row in field window	*/
+  int			curcol;		/* current col in field window	*/
+  int			toprow;		/* in scrollable field window	*/
+  int			begincol;	/* in horiz. scrollable field	*/
+  short			maxfield;	/* number of fields		*/
+  short			maxpage;	/* number of pages		*/
+  short			curpage;	/* index into page		*/
+  Form_Options		opts;		/* options			*/
+  WINDOW *		win;		/* window			*/
+  WINDOW *		sub;		/* subwindow			*/
+  WINDOW *		w;		/* window for current field	*/
+  FIELD **		field;		/* field [maxfield]		*/
+  FIELD *		current;	/* current field		*/
+  _PAGE *		page;		/* page [maxpage]		*/
+  void *		usrptr;		/* user pointer			*/
 
-  void                  (*forminit)(struct formnode *);
-  void                  (*formterm)(struct formnode *);
-  void                  (*fieldinit)(struct formnode *);
-  void                  (*fieldterm)(struct formnode *);
+  void			(*forminit)(struct formnode *);
+  void			(*formterm)(struct formnode *);
+  void			(*fieldinit)(struct formnode *);
+  void			(*fieldterm)(struct formnode *);
 
 } FORM;
 
@@ -164,7 +180,7 @@ typedef void (*Form_Hook)(FORM *);
 #define O_AUTOSKIP		(0x0040)
 #define O_NULLOK		(0x0080)
 #define O_PASSOK		(0x0100)
-#define O_STATIC                (0x0200)
+#define O_STATIC		(0x0200)
 
 /* form options */
 #define O_NL_OVERLOAD		(0x0001)
@@ -220,14 +236,14 @@ typedef void (*Form_Hook)(FORM *);
 #define REQ_SCR_BLINE	 (KEY_MAX + 44)	/* scroll field backward a line	*/
 #define REQ_SCR_FPAGE	 (KEY_MAX + 45)	/* scroll field forward a page	*/
 #define REQ_SCR_BPAGE	 (KEY_MAX + 46)	/* scroll field backward a page	*/
-#define REQ_SCR_FHPAGE   (KEY_MAX + 47) /* scroll field forward  half page */
-#define REQ_SCR_BHPAGE   (KEY_MAX + 48) /* scroll field backward half page */
-#define REQ_SCR_FCHAR    (KEY_MAX + 49) /* horizontal scroll char          */
-#define REQ_SCR_BCHAR    (KEY_MAX + 50) /* horizontal scroll char          */
-#define REQ_SCR_HFLINE   (KEY_MAX + 51) /* horizontal scroll line          */
-#define REQ_SCR_HBLINE   (KEY_MAX + 52) /* horizontal scroll line          */
-#define REQ_SCR_HFHALF   (KEY_MAX + 53) /* horizontal scroll half line     */
-#define REQ_SCR_HBHALF   (KEY_MAX + 54) /* horizontal scroll half line     */
+#define REQ_SCR_FHPAGE	 (KEY_MAX + 47) /* scroll field forward	 half page */
+#define REQ_SCR_BHPAGE	 (KEY_MAX + 48) /* scroll field backward half page */
+#define REQ_SCR_FCHAR	 (KEY_MAX + 49) /* horizontal scroll char	*/
+#define REQ_SCR_BCHAR	 (KEY_MAX + 50) /* horizontal scroll char	*/
+#define REQ_SCR_HFLINE	 (KEY_MAX + 51) /* horizontal scroll line	*/
+#define REQ_SCR_HBLINE	 (KEY_MAX + 52) /* horizontal scroll line	*/
+#define REQ_SCR_HFHALF	 (KEY_MAX + 53) /* horizontal scroll half line	*/
+#define REQ_SCR_HBHALF	 (KEY_MAX + 54) /* horizontal scroll half line	*/
 
 #define REQ_VALIDATION	 (KEY_MAX + 55)	/* validate field		*/
 #define REQ_NEXT_CHOICE	 (KEY_MAX + 56)	/* display next field choice	*/
@@ -256,15 +272,15 @@ extern NCURSES_EXPORT_VAR(FIELDTYPE *) TYPE_INTEGER;
 extern NCURSES_EXPORT_VAR(FIELDTYPE *) TYPE_NUMERIC;
 extern NCURSES_EXPORT_VAR(FIELDTYPE *) TYPE_REGEXP;
 
-        /************************************
+	/************************************
 	*  built-in additional field types  *
-        *  They are not defined in SVr4     *
+	*  They are not defined in SVr4     *
 	************************************/
 extern NCURSES_EXPORT_VAR(FIELDTYPE *) TYPE_IPV4;      /* Internet IP Version 4 address */
 
-        /*********************** 
-        *   Default objects    *
-        ***********************/ 
+	/***********************
+	*   Default objects    *
+	***********************/
 extern NCURSES_EXPORT_VAR(FORM *)	_nc_Default_Form;
 extern NCURSES_EXPORT_VAR(FIELD *)	_nc_Default_Field;
 
@@ -274,8 +290,9 @@ extern NCURSES_EXPORT_VAR(FIELD *)	_nc_Default_Field;
 	***********************/
 extern NCURSES_EXPORT(FIELDTYPE *) new_fieldtype (
 		    bool (* const field_check)(FIELD *,const void *),
-		    bool (* const char_check)(int,const void *)),
-                *link_fieldtype(FIELDTYPE *,FIELDTYPE *);
+		    bool (* const char_check)(int,const void *));
+extern NCURSES_EXPORT(FIELDTYPE *) link_fieldtype(
+		    FIELDTYPE *, FIELDTYPE *);
 
 extern NCURSES_EXPORT(int)	free_fieldtype (FIELDTYPE *);
 extern NCURSES_EXPORT(int)	set_fieldtype_arg (FIELDTYPE *,
@@ -327,7 +344,7 @@ extern NCURSES_EXPORT(FIELDTYPE *)	field_type (const FIELD *);
 
 extern NCURSES_EXPORT(char *)	field_buffer (const FIELD *,int);
 
-extern NCURSES_EXPORT(Field_Options)	field_opts (const FIELD *); 
+extern NCURSES_EXPORT(Field_Options)	field_opts (const FIELD *);
 
 	/******************
 	*  FORM routines  *
